# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
MultiFastaDB -- present a collection of indexed fasta files as a single source

from multifastafile import MultiFastaDB
mff = MultiFastaDB([ '/a/file.fasta', '/a/dir/of/fastas/' ])
mff.fetch('NM_01234.5',60,70)
mff['NM_01234.5'][60:70] (equivalent to the above)
mff.where_is('NM_01234.5')

NOTE: This is temporarily in the HGVS package. It will likely move to a
separate package of bioinformatics tools.
"""


import collections
import itertools
import logging
import os
import re

import pysam

class MultiFastaDB(object):
    """Class that opens a set of indexed fasta files and provides ordered
    lookup of a given accession across all of them.  The intent is to
    simplify accessing a virtual database of sequences that is distributed
    across multiple files."""

    class SequenceProxy(object):

        """Represents a sequence fetch but defers the actual fetch until a
        slice is applied, allowing convenient and sexy slice syntax like this:
        frsf = MultiFastaDB('/my/seqs')
        seq = frsf['NM_01234.5'][4:20]
        """

        def __init__(self,frsf,ac):
            self.frsf = frsf
            self.ac = ac

        def __getslice__(self,start_i,end_i):
            return self.frsf.fetch(self.ac,start_i,end_i)

        def __getitem__(self,i):
            return self[i,i+1]

        def __str__(self):
            return self.frsf.fetch(self.ac)

    # Pysam doesn't support zipped files - returns wrong query results
    # default_suffixes = ['fa','fasta', 'faa', 'fna', 'fa.gz', 'fasta.gz', 'faa.gz', 'fna.gz']

    def __init__(self,sources=[],suffixes=['fa','fasta', 'faa', 'fna'], use_meta_index=False):
        self._fastas = None
        self._index = None
        self._logger = logging.getLogger(__name__)
        self.sources = sources
        self.suffixes = suffixes
        self.use_meta_index = use_meta_index
        self.open_sources()

    def open_sources(self):
        """Opens or reopens fasta sources (directories or files) provided
        when the instance was created."""
        self._fastas = collections.OrderedDict()
        for source in self.sources:
            fa_paths = []
            if os.path.isfile(source):
                fa_paths = [ source ]
            elif os.path.isdir(source):
                # sort so that searches have defined ordering
                fa_paths = sorted([ os.path.join(source,de)
                                    for de in os.listdir(source)
                                    if any(de.endswith(sfx) for sfx in self.suffixes) ])
            else:
                raise RuntimeError(source + ": expected a list of directories or files for fasta sources")
            for fa_path in fa_paths:
                self._logger.debug("opening "+fa_path)
                fai_path = fa_path+'.fai'
                if (os.path.exists(fai_path)
                    and os.stat(fa_path).st_mtime > os.stat(fai_path).st_mtime):
                    self._logger.warn(fai_path + " is out-of-date (older than fasta file)")
                self._fastas[fa_path] = pysam.Fastafile(fa_path)
        self.create_index()

    def create_index(self):
        """
        Create a convenience meta index in the form of an ordered dict that contains just the reference accession and not the full accession
        from the FASTA file
        """
        self._index = collections.OrderedDict()
        ncbi_re = re.compile('ref\|([^|]+)')
        for ref in self.references:
            acs = [ref]
            if self.use_meta_index:
                acs += ncbi_re.findall(ref)
            for ac in acs:
                if ac not in self._index:
                    self._index[ac] = ref
                else:
                    files = [e[0] for e in self.where_is(ac)]
                    self._logger.debug('duplicate sequence found for {ac} in {files}'.format(ac=ac, files=', '.join(files)))

    def where_is(self,ac):
        """return list of all (filename,pysam.Fastafile) pairs in which
        accession occurs"""
        return [ (fp,fh)
                 for fp,fh in self._fastas.iteritems()
                 if ac in fh ]

    def fetch(self,ac,start_i=None,end_i=None):
        """return a sequence, or subsequence if start_i and end_i are provided"""
        for fah in self._fastas.values():
            seq = fah.fetch(self._index[ac],start_i,end_i)
            if seq != '':
                return seq
        return None


    @property
    def references(self):
        return list( itertools.chain.from_iterable([
            fa.references for fa in self._fastas.values() ]))
        
    @property
    def lengths(self):
        return list( itertools.chain.from_iterable([
            fa.lengths for fa in self._fastas.values() ]))


    def __contains__(self,ac):
        return any([ ac in fh for fh in self._fastas.itervalues() ])

    def __getitem__(self,ac):
        return self.SequenceProxy(self,ac)

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

#
# Utility to insert an hgvs variant into a transcript sequence.
# Generates a record corresponding to the modified transcript sequence,
# along with annotations for use in conversion to an hgvsp tag.
# Used in hgvsc to hgvsp conversion.
#
import math
import recordtype

from Bio.Seq import Seq

import hgvs.edit
from hgvs.location import CDS_START, CDS_END

DBG = False

class AltTranscriptData(recordtype.recordtype('AltTranscriptData', [
        'transcript_sequence', 'aa_sequence', 'cds_start', 'cds_stop', 'protein_accession',
        ('is_frameshift', False), ('variant_start_aa', None), ('frameshift_start', None), ('is_substitution', False),
        ('is_ambiguous', False)])):

    @classmethod
    def create_for_variant_inserter(cls, seq, cds_start, cds_stop, is_frameshift, variant_start_aa, accession,
                                    is_substitution=False, is_ambiguous=False):
        """Create a variant sequence using inputs from VariantInserter
        :param seq: DNA sequence wiith variant incorporated
        :type seq: str or list
        :param cds_start: coding sequence start (1-based)
        :type cds_start: int
        :param cds_stop: coding sequence stop (1-based)
        :type cds_stop: int
        :param protein_accession: protein accession, e.g. NP_999999.2
        :type protein_accession: str
        :param is_frameshift: is this variant a frameshift
        :type is_frameshift: bool
        :param variant_start_aa: AA start index (1-based) for this variant
        :type variant_start_aa: int
        :param is_substitution: flag if this is a substitution AA variant
        :type is_substitution: bool
        :param is_ambiguous: flag if variant is "?"
        :type is_ambiguous: bool
        :return variant sequence data
        :rtype recordtype
        """
        if len(seq) > 0:
            if isinstance(seq, basestring):
                seq = list(seq)
            seq_cds = seq[cds_start - 1:]
            if len(seq_cds) % 3 != 0:   # padding so biopython won't complain during the conversion
                seq_cds.extend(['N']*((3-len(seq_cds) % 3) % 3))
            seq_cds = ''.join(seq_cds)
            seq_aa = str(Seq(seq_cds).translate())
            stop_pos = seq_aa.find("*")
            if stop_pos != -1:
                seq_aa = seq_aa[:stop_pos + 1]
        else:
            seq_aa = []

        alt_data = AltTranscriptData(''.join(seq), seq_aa, cds_start, cds_stop, accession,
                                     is_frameshift=is_frameshift, variant_start_aa=variant_start_aa,
                                     is_substitution=is_substitution, is_ambiguous=is_ambiguous)

        return alt_data


class AltSeqBuilder(object):

    EXON = "exon"
    INTRON = "intron"
    F_UTR = "five utr"
    T_UTR = "three utr"
    WHOLE_GENE = "whole gene"

    def __init__(self, var_c, transcript_data):
        """Constructor

        :param var_c: representation of hgvs variant
        :type var_c: SequenceVariant
        :param transcript_data: representation of transcript
        :type transcript_data: recordtype

        """
        self._var_c = var_c
        self._transcript_data = transcript_data
        if DBG:
            print( transcript_data.transcript_sequence )

        # check reference for special characteristics
        self._ref_has_multiple_stops = self._transcript_data.aa_sequence.count("*") > 1

    def build_altseq(self):
        """given a variant and a sequence, incorporate the variant and return the new sequence

        Data structure returned is analogous to the data structure used to return the variant sequence,
        but with an additional parameter denoting the start of a frameshift that should affect all bases
        downstream.

        :returns variant sequence data
        :rtype list of dictionaries
        """
        NOT_CDS = "not_cds_variant"
        WHOLE_GENE_DELETED = "whole_gene_deleted"

        type_map = {hgvs.edit.NARefAlt: self._incorporate_delins,
                    hgvs.edit.Dup: self._incorporate_dup,
                    hgvs.edit.Repeat: self._incorporate_repeat,
                    NOT_CDS: self._create_alt_equals_ref_noncds,
                    WHOLE_GENE_DELETED: self._create_no_protein
                    }

        # TODO - loop over each allele rather than assume only 1 variant; return a list for now
        alt_data = []

        variant_location = self._get_variant_region()

        if variant_location == self.EXON:
            edit_type = type(self._var_c.posedit.edit)
        elif variant_location == self.INTRON:
            edit_type = NOT_CDS
        elif variant_location == self.T_UTR:
            edit_type = NOT_CDS
        elif variant_location == self.F_UTR:
            # TODO - handle case where variant introduces a Met (new start)
            edit_type = NOT_CDS
        elif variant_location == self.WHOLE_GENE:
            if self._var_c.posedit.edit.ref is not None and self._var_c.posedit.edit.alt is None:
                edit_type = WHOLE_GENE_DELETED
            else:
                edit_type = NOT_CDS
        else:   # should never get here
            raise ValueError("value_location = {}".format(variant_location))

        try:
            this_alt_data = type_map[edit_type]()
        except KeyError as e:
            raise NotImplementedError("c to p translation unsupported for {} type {}".format(self._var_c, edit_type))

        # get the start of the "terminal" frameshift (i.e. one never "cancelled out")
        this_alt_data = self._get_frameshift_start(this_alt_data)
        alt_data.append(this_alt_data)
        if DBG:
            print( this_alt_data.transcript_sequence )

        return alt_data

    def _get_variant_region(self):
        """Categorize variant by location in transcript (5'utr, exon, intron, 3'utr)

        :return "exon", "intron", "five_utr", "three_utr", "whole_gene"
        :rtype str
        """
        if self._var_c.posedit.pos.start.datum == CDS_END and self._var_c.posedit.pos.end.datum == CDS_END:
            result = self.T_UTR
        elif self._var_c.posedit.pos.start.base < 0 and self._var_c.posedit.pos.end.base < 0:
            result = self.F_UTR
        elif self._var_c.posedit.pos.start.base < 0 and self._var_c.posedit.pos.end.datum == CDS_END:
            result = self.WHOLE_GENE
        elif self._var_c.posedit.pos.start.offset != 0 or self._var_c.posedit.pos.end.offset != 0:
            # leave out anything intronic for now
            result = self.INTRON
        else:   # anything else that contains an exon
            result = self.EXON
        return result

    # def _is_intron_only(self):
    #     """Checks if variant is entirely intronic"""
    #
    #     # case 1: same base,same sign on offset -> start, end are anchored from the same base
    #     same_start = self._var_c.posedit.pos.start.base == self._var_c.posedit.pos.end.base
    #     pos_offset = self._var_c.posedit.pos.start.offset > 0 and self._var_c.posedit.pos.end.offset > 0
    #     neg_offset = self._var_c.posedit.pos.start.offset < 0 and self._var_c.posedit.pos.end.offset < 0
    #
    #     # case 2: start, end between 2 different exon bases but don't overlap either
    #     same_start_plus_1 = (self._var_c.posedit.pos.start.base + 1) == self._var_c.posedit.pos.end.base
    #     pm_offset = self._var_c.posedit.pos.start.offset > 0 and self._var_c.posedit.pos.end.offset < 0
    #
    #     # other types of introns would overlap at least 1 exon base
    #
    #     return (same_start and (pos_offset or neg_offset)) or (same_start_plus_1 and pm_offset)


    def _incorporate_delins(self):
        """Incorporate delins"""
        seq, cds_start, cds_stop, start, end = self._setup_incorporate()

        ref = self._var_c.posedit.edit.ref
        alt = self._var_c.posedit.edit.alt
        ref_length = end - start if ref is not None else 0  # can't just get from ref since ref isn't always known
        alt_length = len(self._var_c.posedit.edit.alt) if self._var_c.posedit.edit.alt is not None else 0
        net_base_change = alt_length - ref_length
        cds_stop += net_base_change

        # incorporate the variant into the sequence (depending on the type)
        is_substitution = False
        if ref is not None and alt is not None:     # delins or SNP
            seq[start:end] = list(alt)
            if len(ref) == 1 and len(alt) == 1:
                is_substitution = True
        elif ref is not None:                       # deletion
            del seq[start:end]
        else:                                       # insertion
            seq[start + 1:start + 1] = list(alt)    # insertion in list before python list index

        if DBG:
            print( "net base change: {}".format(net_base_change) )
        is_frameshift = net_base_change % 3 != 0
        # use max of mod 3 value and 1 (in event that indel starts in the 5'utr range)
        variant_start_aa = max(int(math.ceil((self._var_c.posedit.pos.start.base) / 3.0)), 1)

        alt_data = AltTranscriptData.create_for_variant_inserter(seq, cds_start, cds_stop,
                                                                 is_frameshift, variant_start_aa,
                                                                 self._transcript_data.protein_accession,
                                                                 is_substitution=is_substitution,
                                                                 is_ambiguous=self._ref_has_multiple_stops)
        return alt_data

    def _incorporate_dup(self):
        """Incorporate dup into sequence"""
        seq, cds_start, cds_stop, start, end = self._setup_incorporate()

        dup_seq = seq[start:end]
        seq[end:end] = dup_seq
        
        is_frameshift = len(dup_seq) % 3 != 0
        variant_start_aa = int(math.ceil((self._var_c.posedit.pos.end.base + 1) / 3.0))

        alt_data = AltTranscriptData.create_for_variant_inserter(seq, cds_start, cds_stop,
                                                                 is_frameshift, variant_start_aa,
                                                                 self._transcript_data.protein_accession,
                                                                 is_ambiguous=self._ref_has_multiple_stops)
        return alt_data

    def _incorporate_repeat(self):
        """Incorporate repeat int sequence"""
        # TODO - implement
        raise NotImplementedError("hgvs c to p conversion does not support {} type: repeats".format(self._var_c))

    def _setup_incorporate(self):
        """Helper to setup incorporate functions
        :return (transcript sequence, cds start [1-based], cds stop [1-based],
        cds start index in seq [inc, 0-based], cds end index in seq [excl, 0-based])
        :rtype (list, int, int, int, int)
        """
        seq = list(self._transcript_data.transcript_sequence)

        # get initial start/end points; will modify these based on the variant length
        cds_start = self._transcript_data.cds_start
        cds_stop = self._transcript_data.cds_stop

        start_end = []
        for pos in (self._var_c.posedit.pos.start, self._var_c.posedit.pos.end):
            # list is zero-based; seq pos is 1-based
            if pos.datum == CDS_START:
                if pos.base < 0:                  # 5' UTR
                    result = cds_start - 1
                else:                             # cds/intron
                    if pos.offset <= 0:
                        result = (cds_start - 1) + pos.base - 1
                    else:
                        result = (cds_start - 1) + pos.base
            elif pos.datum == CDS_END:            # 3' UTR
                result = cds_stop + pos.base - 1
            else:
                raise NotImplementedError("Unsupported/unexpected location")
            start_end.append(result)

        # unpack; increment end by 1 (0-based exclusive)
        (start, end) = start_end
        end += 1

        if DBG:
            print( "len seq:{} cds_start:{} cds_stop:{} start:{} end:{}".format(len(seq), cds_start, cds_stop, start, end) )
        return seq, cds_start, cds_stop, start, end

    def _create_alt_equals_ref_noncds(self):
        """Create an alt seq that matches the reference (for non-cds variants)"""
        alt_data = AltTranscriptData.create_for_variant_inserter(list(self._transcript_data.transcript_sequence),
                                                                 self._transcript_data.cds_start,
                                                                 self._transcript_data.cds_stop,
                                                                 False,
                                                                 None,
                                                                 self._transcript_data.protein_accession,
                                                                 is_ambiguous=True)
        return alt_data

    def _create_no_protein(self):
        """Create a no-protein result"""
        alt_data = AltTranscriptData.create_for_variant_inserter([],
                                                                 None,
                                                                 None,
                                                                 False,
                                                                 None,
                                                                 self._transcript_data.protein_accession,
                                                                 is_ambiguous=False)
        return alt_data

    def _get_frameshift_start(self, variant_data):
        """Get starting position (AA ref index) of the last frameshift
        which affects the rest of the sequence, i.e. not offset by subsequent frameshifts
        :param variant_data: info on each variant
        :type variant_data: recordtype
        :return variant data with additional field for AA index (1-based) of the frameshift start
        :rtype recordtype
        """
        # TODO - implement for 2+ variants

        if DBG:
            print( "is_frameshift:{}".format(variant_data.is_frameshift) )
            print( "variant_start_aa:{}".format(variant_data.variant_start_aa) )
        if variant_data.is_frameshift:
            variant_data.frameshift_start = variant_data.variant_start_aa
        return variant_data


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

from hgvs.dataproviders.multifastadb import MultiFastaDB
from hgvs.exceptions import HGVSValidationError
import hgvs.dataproviders.uta
import hgvs.variantmapper
import hgvs.parser
import hgvs.validator

db_dir = ['tests/data/sample_data']
hdp = hgvs.dataproviders.uta.connect()
mfdb = MultiFastaDB(db_dir, use_meta_index=True)


class Test_HGVSValidator(unittest.TestCase):
    """Validator wrapper class tests (most testing is handled by the component classes)"""

    def setUp(self):
        self.hp = hgvs.parser.Parser()
        self.vr = hgvs.validator.Validator(hdp,mfdb)

    def test_wrapper(self):
        """Test that validator wrapper is working"""
        self.assertTrue(self.vr.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.6C>A')))


@attr(tags=["quick","validation"])
class Test_HGVSIntrinsicValidator(unittest.TestCase):
    """Tests for internal validation"""

    def setUp(self):
        self.hp = hgvs.parser.Parser()
        self.validate_int = hgvs.validator.IntrinsicValidator()

    def test_start_lte_end(self):
        """Test if start position is less <= end position"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662C>T')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662_36561663insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NM_01234.5:c.76_77insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+55insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54A>T')))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.BASE_RANGE_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561664_36561663A>T'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.BASE_RANGE_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('NM_000277.1:c.*1_2delAG'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.OFFSET_RANGE_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+56_123+55A>T'))

    def test_ins_length_is_one(self):
        """Test if insertion length = 1"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662_36561663insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_77insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+55insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123-54_123-53insT')))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.INS_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78insTT'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.INS_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+56insT'))

    def test_del_length(self):
        """Test if del length agrees with position range"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78delACT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+55delTA')))  # <-- haha "delta"

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.DEL_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78del'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.DEL_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78delACTACAT'))

    def test_sub(self):
        """Test substitution ref != alt"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662C>T')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54A>T')))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.SUB_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662_36561663T>T'))


@attr(tags=["validation"])
class Test_HGVSExtrinsicValidator(unittest.TestCase):
    """Tests for external validation"""

    def setUp(self):
        self.hp = hgvs.parser.Parser()
        self.validate_ext = hgvs.validator.ExtrinsicValidator(hdp,mfdb)

    def test_valid_ref(self):
        """Test if reference seqeuence is valid. Uses sample_data in tests directory"""
        self.assertTrue(self.validate_ext.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.6C>A')))
        self.assertTrue(self.validate_ext.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.-38T>A')))
        self.assertTrue(self.validate_ext.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.*3C>G')))
        self.assertTrue(self.validate_ext.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.435_440delCTGCTG')))
        #self.assertTrue(self.validate_ext.validate(self.hp.parse_hgvs_variant('NP_001005405.1:p.Gly2Ser')))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.SEQ_ERROR_MSG):
            self.validate_ext.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.435_440delCTGCT'))


if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

import hgvs.edit
from hgvs.exceptions import HGVSError

@attr(tags=["quick","models"])
class Test_Edit(unittest.TestCase):
    def test_NARefAlt_exceptions(self):
        with self.assertRaises(HGVSError):
            edit = str(hgvs.edit.NARefAlt(None,None))

    def test_NARefAlt(self):
        self.assertEqual( str(hgvs.edit.NARefAlt('A','A'  ))  , '='             )
        self.assertEqual( str(hgvs.edit.NARefAlt('A','T'  ))  , 'A>T'           )
        self.assertEqual( str(hgvs.edit.NARefAlt('AA',None))  , 'delAA'         )
        self.assertEqual( str(hgvs.edit.NARefAlt(None,'TT'))  , 'insTT'         )
        self.assertEqual( str(hgvs.edit.NARefAlt('AA','T' ))  , 'delAAinsT'     )
        self.assertEqual( str(hgvs.edit.NARefAlt('A','TT' ))  , 'delAinsTT'     )
        # edit types
        self.assertEqual( str(hgvs.edit.NARefAlt('A','A'  ).type)  , 'identity' )
        self.assertEqual( str(hgvs.edit.NARefAlt('A','T'  ).type)  , 'sub'      )
        self.assertEqual( str(hgvs.edit.NARefAlt('AA',None).type)  , 'del'      )
        self.assertEqual( str(hgvs.edit.NARefAlt(None,'TT').type)  , 'ins'      )
        self.assertEqual( str(hgvs.edit.NARefAlt('AA','T' ).type)  , 'delins'   )
        self.assertEqual( str(hgvs.edit.NARefAlt('A','TT' ).type)  , 'delins'   )

    def test_AARefAlt(self):
        self.assertEqual( str(hgvs.edit.AARefAlt('A','A'  ))  , '='             )
        self.assertEqual( str(hgvs.edit.AARefAlt('A','T'  ))  , 'Thr'           )
        self.assertEqual( str(hgvs.edit.AARefAlt('AA',None))  , 'del'           )
        self.assertEqual( str(hgvs.edit.AARefAlt(None,'TT'))  , 'insThrThr'     )
        self.assertEqual( str(hgvs.edit.AARefAlt('','T' ))  ,   'delinsThr'     )
        self.assertEqual( str(hgvs.edit.AARefAlt('AA','T' ))  , 'delinsThr'     )
        self.assertEqual( str(hgvs.edit.AARefAlt('A','TT' ))  , 'delinsThrThr'  )
        # edit types
        self.assertEqual( str(hgvs.edit.AARefAlt('A','A'  ).type)  , 'identity' )
        self.assertEqual( str(hgvs.edit.AARefAlt('A','T'  ).type)  , 'sub'      )
        self.assertEqual( str(hgvs.edit.AARefAlt('AA',None).type)  , 'del'      )
        self.assertEqual( str(hgvs.edit.AARefAlt(None,'TT').type)  , 'ins'      )
        self.assertEqual( str(hgvs.edit.AARefAlt('','T' ).type)  ,   'delins'   )
        self.assertEqual( str(hgvs.edit.AARefAlt('AA','T' ).type)  , 'delins'   )
        self.assertEqual( str(hgvs.edit.AARefAlt('A','TT' ).type)  , 'delins'   )

    def test_AASub(self):
        self.assertEqual( str(hgvs.edit.AASub('A','T'  ))  , 'Thr'           )
        # edit types
        self.assertEqual( str(hgvs.edit.AASub('A','T'  ).type)  , 'sub'      )

    def test_AAFs(self):
        self.assertEqual( str(hgvs.edit.AAFs('A','T','6'))  , 'ThrfsTer6'  )
        self.assertEqual( str(hgvs.edit.AAFs('A','T','?'))  , 'ThrfsTer?'  )
        self.assertEqual( str(hgvs.edit.AAFs('A','T', None))  , 'ThrfsTer' )
        # edit types
        self.assertEqual( str(hgvs.edit.AAFs('A','T','6').type)  , 'fs'    )
        self.assertEqual( str(hgvs.edit.AAFs('A','T','?').type)  , 'fs'    )
        self.assertEqual( str(hgvs.edit.AAFs('A','T', None).type)  , 'fs'  )

    def test_AAExt(self):
        self.assertEqual( str(hgvs.edit.AAExt('A','V', '*', 10))  , 'ValextTer10'  )
        self.assertEqual( str(hgvs.edit.AAExt('A','V', None, -10))  , 'Valext-10'  )
        self.assertEqual( str(hgvs.edit.AAExt('A',None, None, -5))  , 'ext-5'      )
        # edit types
        self.assertEqual( str(hgvs.edit.AAExt('A','V', '*', 10).type)  , 'ext'     )
        self.assertEqual( str(hgvs.edit.AAExt('A','V', None, -10).type)  , 'ext'   )
        self.assertEqual( str(hgvs.edit.AAExt('A',None, None, -5).type)  , 'ext'   )

    def test_Dup(self):
        self.assertEqual( str(hgvs.edit.Dup())				, 'dup' 		)
        self.assertEqual( str(hgvs.edit.Dup('T'))			, 'dupT' 		)
        # edit types
        self.assertEqual( str(hgvs.edit.Dup().type)				, 'dup' 		)
        self.assertEqual( str(hgvs.edit.Dup('T').type)			, 'dup' 		)

    def test_Repeat(self):
        self.assertEqual( str(hgvs.edit.Repeat('CAG',12,34)), 'CAG(12_34)' 	)
        # edit types
        self.assertEqual( str(hgvs.edit.Repeat('CAG',12,34).type), 'repeat' 	)

    def test_Repeat_exceptions(self):
        with self.assertRaises(HGVSError):
            edit = str(hgvs.edit.Repeat('CAG',34,12))

if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import unittest

import hgvs.dataproviders.uta

class UTA_Base(object):

    def test_get_acs_for_protein_seq(self):
        exp = ['NP_001005405.1','MD5_8fc09b1d9a38a8c55176a0fa922df227']
        s = """
        mgccgcsggc gsgcggcgsg sggcgsgcgg cgssccvpic cckpvcccvp acscsscgsc
        ggskggcgsc gsskggcgsc gcsqsncckp ccsssgcgsf ccqsscskpc ccqssccqss
        cckpcccqss ccqsscfkpc ccqssccvpv ccqcki
        """
        s = re.sub('\s+','',s.upper())
        self.assertEqual( sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp) )

        exp = ['NP_071928.2','MD5_ffb0d4adbd5e0b5d71678228b3696984']
        s = """
        masetektha llqtcstesl isslglgafc lvadrllqfs tiqqndwlra lsdnavhcvi
        gmwswavvtg ikkktdfgei ilagflasvi dvdhfflags mslkaaltlp rrpflhcstv
        ipvvvltlkf tmhlfklkds wcflpwmlfi swtshhirdg irhglwicpf gktsplpfwl
        yviitsslph icsfvmyltg trqmmsskhg vridv
        """
        s = re.sub('\s+','',s.upper())
        self.assertEqual( sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp) )

    def test_get_gene_info(self):
        gene_info = self.hdp.get_gene_info('VHL')
        self.assertEqual( 'VHL', gene_info['hgnc'])
        self.assertEqual( '3p25.3', gene_info['maploc'])
        self.assertEqual( 6, len(gene_info))

    def test_get_tx_exons(self):    
        tx_exons = self.hdp.get_tx_exons('NM_000551.3','NC_000003.11', 'splign')
        self.assertEqual( 3, len(tx_exons) )

    def test_get_tx_exons_invalid_tx_ac(self):
        tx_exons = self.hdp.get_tx_exons('NM_999999.9','NC_000003.11', 'splign')
        self.assertIsNone( tx_exons )

    def test_get_tx_exons_invalid_alt_ac(self):
        tx_exons = self.hdp.get_tx_exons('NM_000551.3','NC_000999.9', 'splign')
        self.assertIsNone( tx_exons )

    def test_get_tx_exons_invalid_alt_aln_method(self):
        tx_exons = self.hdp.get_tx_exons('NM_000551.3','NC_000999.9', 'best')
        self.assertIsNone( tx_exons )

    def test_get_tx_for_gene(self):
        tig = self.hdp.get_tx_for_gene('VHL')
        self.assertEqual( 11, len(tig) )

    def test_get_tx_for_gene_invalid_gene(self):
        tig = self.hdp.get_tx_for_gene('GENE')
        self.assertEqual( 0, len(tig) )


    def test_get_tx_info(self):
        tx_info = self.hdp.get_tx_info('NM_000051.3', 'AC_000143.1', 'splign')
        self.assertEqual( 385, tx_info['cds_start_i'])
        self.assertEqual( 9556, tx_info['cds_end_i'])
        self.assertEqual( 'AC_000143.1', tx_info['alt_ac'])

    def test_get_tx_info_invalid_tx_ac(self):
        tx_info = self.hdp.get_tx_info('NM_999999.9', 'AC_000143.1', 'splign')
        self.assertIsNone( tx_info )

    def test_get_tx_mapping_options(self):
        tx_mapping_options = self.hdp.get_tx_mapping_options('NM_000551.3')
        self.assertEqual( 4, len(tx_mapping_options) )
        self.assertIn( ['NM_000551.3', 'NC_000003.11', 'splign'], tx_mapping_options )
        self.assertIn( ['NM_000551.3', 'NC_000003.11', 'blat']  , tx_mapping_options )

    def test_get_tx_mapping_options_invalid(self):
        tx_info_options = self.hdp.get_tx_mapping_options('NM_999999.9')
        self.assertEqual( tx_info_options, [] )

    def test_get_tx_seq(self):
        tsi = self.hdp.get_tx_seq('NM_000551.3')
        seq = tsi
        self.assertEqual( 4560, len(seq) )
        self.assertTrue( seq.startswith('cctcgcctccgttacaacgg'.upper()) )
        self.assertTrue( seq.endswith('aaaggaaatag'.upper()) )

    def test_get_tx_seq_invalid_ac(self):
        tsi = self.hdp.get_tx_seq('NM_999999.9')
        self.assertIsNone( tsi )



class Test_hgvs_dataproviders_uta_UTA_default(unittest.TestCase, UTA_Base):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect()
        return self


if __name__ == '__main__':
    unittest.main()
        

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

import hgvs.edit
import hgvs.location
import hgvs.posedit

@attr(tags=["quick","models","models"])
class Test_PosEdit(unittest.TestCase):
    def test_PosEdit(self):
        pos = hgvs.location.Interval(
                hgvs.location.BaseOffsetPosition(base=12,offset=+34),
                hgvs.location.BaseOffsetPosition(base=56,offset=-78) )
        edit = hgvs.edit.NARefAlt('AA',None)
        pe = hgvs.posedit.PosEdit( pos=pos, edit=edit )
        self.assertEqual( str(pe), '12+34_56-78delAA' )

if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import re

def strand_pm(s):
    return (None if s is None
            else '+' if s == 1
            else '-' if s == -1
            else '?')


def build_tx_cigar(exons, strand):
    cigarelem_re = re.compile('\d+[=DIMNX]')
    def _reverse_cigar(c):
        return ''.join(reversed(cigarelem_re.findall(c)))

    if len(exons) == 0:
        return None

    if strand == -1:
        for i in range(len(exons)):
            exons[i]['cigar'] = _reverse_cigar(exons[i]['cigar'])

    tx_cigar = [exons[0]['cigar']]    # exon 1
    for i in range(1, len(exons)):    # and intron + exon pairs thereafter
        tx_cigar += [str(exons[i]['alt_start_i'] - exons[i-1]['alt_end_i']) + 'N',
                     exons[i]['cigar']]
    
    tx_cigar_str = ''.join(tx_cigar)
    
    return tx_cigar_str





## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import string

complement_transtable_str = string.maketrans('ACGT','TGCA')
complement_transtable_uni = dict(zip(map(ord,u'ACGT'),u'TGCA'))
def reverse_complement(s):
    def _rc_str(s):
        return ''.join(reversed(s.translate(complement_transtable_str)))
    def _rc_uni(s):
        return ''.join(reversed(s.translate(complement_transtable_uni)))
    return None if s is None else _rc_uni(s) if isinstance(s,unicode) else _rc_str(s)




## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import abc

class Interface(object):
    """Variant mapping and validation requires access to external data,
    specifically exon structures, transcript alignments, and protein
    accessions.  In order to isolate the hgvs package from the myriad
    choices and tradeoffs, these data are provided through an
    implementation of the (abstract) HGVS Data Provider Interface.
    
    As of June 2014, the only available data provider implementation
    uses the Universal Transcript Archive (`UTA`_), a sister project
    that provides access to transcripts and genome-transcript
    alignments.  `Invitae`_ provides a public UTA database instance
    that is used by default; see the `UTA`_ page for instructions on
    installing your own PostgreSQL or SQLite version.  In the future,
    other implementations may be availablefor other data sources.

    Pure virtural class for the HGVS Data Provider Interface.  Every
    data provider implementation should be a subclass (possibly
    indirect) of this class.

    .. _UTA: http://bitbucket.org/invitae/uta
    .. _Invitae: http://invitae.com/
    """

    __metaclass__ = abc.ABCMeta

    def interface_version(self):
        return 1

    @abc.abstractmethod
    def data_version(self):
        pass

    @abc.abstractmethod
    def schema_version(self):
        pass

    @abc.abstractmethod
    def get_tx_exons(self, tx_ac, alt_ac, alt_aln_method):
        pass

    @abc.abstractmethod
    def get_tx_info(self, tx_ac, alt_ac, alt_aln_method):
        pass

    @abc.abstractmethod
    def get_tx_seq(self,ac):
        pass

    @abc.abstractmethod
    def get_tx_for_gene(self, gene):
        pass

    @abc.abstractmethod
    def get_acs_for_protein_seq(seq):
        pass

    @abc.abstractmethod
    def get_gene_info(self, gene):
        pass

    @abc.abstractmethod
    def get_tx_mapping_options(self, tx_ac):
        pass

    @abc.abstractmethod
    def get_tx_identity_info(self, tx_ac):
        pass

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
hgvs.hgvsvalidator
"""

from hgvs.exceptions import HGVSValidationError, HGVSDataNotAvailableError
import hgvs.variantmapper
import hgvs.parser


BASE_RANGE_ERROR_MSG = 'base start position must be <= end position'
OFFSET_RANGE_ERROR_MSG = 'offset start must be <= end position'
INS_ERROR_MSG = 'insertion length must be 1'
DEL_ERROR_MSG = 'start and end position range must equal sequence deletion length'
SUB_ERROR_MSG = 'substitution cannot have the same ref and alt base'
AC_ERROR_MSG = 'Accession is not present in BDI database'
SEQ_ERROR_MSG = 'Variant reference does not agree with reference sequence'



class Validator(object):
    """invoke intrinsic and extrinsic validation"""
    def __init__(self, hdp, mfdb=None):
        self.ivr = IntrinsicValidator()
        self.evr = ExtrinsicValidator(hdp,mfdb)

    def validate(self,var):
        return self.ivr.validate(var) and self.evr.validate(var)


class IntrinsicValidator(object):
    """
    Attempts to determine if the HGVS name is internally consistent
    """

    def validate(self, var):
        assert isinstance(var, hgvs.variant.SequenceVariant), 'variant must be a parsed HGVS sequence variant object'
        self._start_lte_end(var)
        self._ins_length_is_one(var)
        self._del_length(var)
        self._sub_alt_is_not_ref(var)
        return True

    def _start_lte_end(self, var):
        if var.type == 'g':
            if var.posedit.pos.start.base > var.posedit.pos.end.base:
                raise HGVSValidationError(BASE_RANGE_ERROR_MSG)
        if var.type in ['c', 'r', 'p']:
            if var.posedit.pos.start.base > var.posedit.pos.end.base:
                raise HGVSValidationError(BASE_RANGE_ERROR_MSG)
            elif var.posedit.pos.start.base == var.posedit.pos.end.base:
                if var.posedit.pos.start.offset > var.posedit.pos.end.offset:
                    raise HGVSValidationError(OFFSET_RANGE_ERROR_MSG)
            if var.posedit.pos.start.datum > var.posedit.pos.end.datum:
                raise HGVSValidationError(BASE_RANGE_ERROR_MSG)
        return True
    
    def _ins_length_is_one(self, var):
        if var.posedit.edit.type == 'ins':
            if var.type == 'g':
                if (var.posedit.pos.end.base - var.posedit.pos.start.base) != 1:
                    raise HGVSValidationError(INS_ERROR_MSG)
            if var.type in ['c', 'r', 'p']:
                if ((var.posedit.pos.end.base + var.posedit.pos.end.offset) -
                        (var.posedit.pos.start.base + var.posedit.pos.start.offset)) != 1:
                    raise HGVSValidationError(INS_ERROR_MSG)
            return True
        
    def _del_length(self, var):
        if var.posedit.edit.type == 'del':
            del_len = len(var.posedit.edit.ref)
            if var.type == 'g':
                if (var.posedit.pos.end.base - var.posedit.pos.start.base + 1) != del_len:
                    raise HGVSValidationError(DEL_ERROR_MSG)
            if var.type in ['c', 'r', 'p']:
                if ((var.posedit.pos.end.base + var.posedit.pos.end.offset) -
                        (var.posedit.pos.start.base + var.posedit.pos.start.offset) + 1) != del_len:
                    raise HGVSValidationError(DEL_ERROR_MSG)
            return True

    def _sub_alt_is_not_ref(self, var):
        if var.posedit.edit.type == 'identity':
            if var.posedit.edit.ref == var.posedit.edit.alt:
                raise HGVSValidationError(SUB_ERROR_MSG)


class ExtrinsicValidator():
    """
    Attempts to determine if the HGVS name validates against external data sources
    """

    def __init__(self, hdp, mfdb=None):
        self.hdp = hdp
        self.mfdb = mfdb
        self.hm = hgvs.variantmapper.VariantMapper(self.hdp)

    def validate(self, var):
        assert isinstance(var, hgvs.variant.SequenceVariant), 'variant must be a parsed HGVS sequence variant object'
        self._ref_is_valid(var)
        return True

    def _ref_is_valid(self, var):
        if var.posedit.edit.ref is not None:
            var_x = self.hm.c_to_r(var) if var.type == 'c' else var
            # fetch uses interbase coordinates
            seq = self._fetch_seq(var_x.ac, var_x.posedit.pos.start.base - 1, var_x.posedit.pos.end.base)
            if seq != var_x.posedit.edit.ref:
                raise HGVSValidationError(str(var) + ': ' + SEQ_ERROR_MSG)
        return True

    def _fetch_seq(self,ac,start_i,end_i):
        """fetch 0-based, right-open subsequence [start_i,end_i) of specified accession
        first using the hdpi instance then mfdb if not None.

        :raises: hgvs.exceptions.HGVSDataNotAvailableError if couldn't get sequence
        """
        q = self.hdp.get_tx_seq(ac)
        if q:
            return q[start_i:end_i]
        if self.mfdb:
            return self.mfdb(ac,start_i,end_i)
        raise HGVSDataNotAvailableError("No sequence available for {ac}".format(ac=ac))




if __name__ == '__main__':
    hgvsparser = hgvs.parser.Parser()
    var1 = hgvsparser.parse_hgvs_variant('NM_001005405.2:r.2T>A')
    validate_ext = ExtrinsicValidator()
    validate_ext.validate(var1)


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/hgvs/hgvs)
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>
